/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;

/**
    This renderer takes as its input a {@link Scene} data structure
    and a {@link FrameBuffer} data structure. This renderer mutates
    the {@link FrameBuffer}'s current viewport so that it is filled
    in with the rendered image of the scene represented by the
    {@link Scene} object.
<p>
    This implements our second rendering pipeline. It has three
    pipeline stages, just like the first pipeline, but the third
    stage, rasterization, now has two versions. One version is
    the same as in the first pipeline, and the second version does
    line clipping at the same time as it does rasterization.
*/
public class Pipeline
{
   public static boolean debug = false;
   public static boolean doClipping = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // Render every Model in the Scene.
      for (Model model : scene.modelList)
      {
         if ( ! model.hidden )
         {
            // 0. Make a deep copy of the model.
            Model model2 = new Model(model);

            // 1. Apply the projection transformation.
            Projection.project(model2.vertexList, scene.camera);

            // 2. Apply the viewport transformation.
            Viewport.viewport(model2.vertexList, fb);

            // 3. Rasterize each line segment into pixels.
            for (LineSegment ls : model2.lineSegmentList)
            {
               if (! doClipping)
                  Rasterize.rasterize(ls, fb);
               else // Rasterize and clip at the same time.
                  Rasterize_Clip.rasterize(ls, fb);
            }
         }
      }
   }



   /**
      Render the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // Render every Model in the Scene.
      for (Model model : scene.modelList)
      {
         System.out.println("=== Render Model ===================================");

         if ( ! model.hidden )
         {
            // 0. Make a deep copy of the model.
            Model model2 = new Model(model);

            // log interesting information to standard output
            logVertexList("0. Input    ", model2.vertexList);

            // 1. Apply the projection transformation.
            Projection.project(model2.vertexList, scene.camera);

            // log interesting information to standard output
            logVertexList("1. Projected", model2.vertexList);

            // 2. Apply the viewport transformation.
            Viewport.viewport(model2.vertexList, fb);

            // log interesting information to standard output
            logVertexList("2. Viewport ", model2.vertexList);

            for (LineSegment ls : model2.lineSegmentList)
            {
               // log interesting information to standard output
               logLineSegment("3. Rasterize", ls);

               // 3. Rasterize the line segment into pixels.
               Rasterize.rasterize(ls, fb);
               // Or, rasterize and clip the line segment into pixels.
             //Rasterize_Clip.rasterize(ls, model2, fb);
            }
            System.out.println();
         }
         else
         {
            System.out.println("   Hidden model.");
         }
         System.out.println("=== End Model ===================================");
      }
      System.out.println("=== End Scene ===================================");
   }


   /**
      This method prints a {@link String} representation of the given vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index = %3d, " + v.toString(), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   private static void logLineSegment(String stage, LineSegment ls)
   {
      System.out.print( stage + ": " + ls.toString() );
   }
}
