/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.awt.Color;

/**
   Rasterize a {@link LineSegment} into pixels in the
   {@link FrameBuffer}'s viewport, but do not rasterize
   any part of the line segment that is not contained in
   the {@link Camera}'s standard view rectangle.
<p>
   This pipeline stage takes a {@link LineSegment} with vertices
   in the logical viewport coordinate system and rasterizezs (into
   pixels in the {@link FrameBuffer}'s viewport) only that part of
   the line segment which is contained in the {@link Camera}'s
   standard view rectangle.
*/
public class Rasterize_Clip
{
   public static boolean debug = false;

   /**
      Rasterize a {@link LineSegment} into pixels in the
      {@link FrameBuffer}'s viewport.

      @param ls  {@link LineSegment} to rasterize into the {@link FrameBuffer}
      @param fb  {@link FrameBuffer} to hold rasterized pixels
   */
   public static void rasterize(LineSegment ls, FrameBuffer fb)
   {
      Color c = Color.white;

      // Make local copies of several values.
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      Vertex v0 = ls.model.vertexList.get( ls.index[0] );
      Vertex v1 = ls.model.vertexList.get( ls.index[1] );

      // Round the line segment's two endpoints to the nearest
      // logical pixel. This makes the algorithm a lot simpler,
      // but it can cause a slight, but noticable, shift of the
      // line segment.
      double x0 = Math.round(v0.x),  y0 = Math.round(v0.y);
      double x1 = Math.round(v1.x),  y1 = Math.round(v1.y);

      // Rasterize a degenerate line segment (a line segment
      // that projected onto a point) as a single pixel.
      if ( (x0 == x1) && (y0 == y1) )
      {
         // We don't know which endpoint of the line segment
         // is in front, so just pick v0.
         fb.setPixelVP((int)x0 - 1, h - (int)y0, c);
         return;
      }

      if (Math.abs(y1 - y0) <= Math.abs(x1 - x0)) // if abs(slope) <= 1
      {
         if (x1 < x0) // swap (x0, y0) with (x1, y1)
         {
            double tempX = x0;
            double tempY = y0;
            x0 = x1;
            y0 = y1;
            x1 = tempX;
            y1 = tempY;
         }

         // Compute this line segment's slope.
         double m = (y1 - y0) / (x1 - x0);

         // Rasterize this line segment.
         // In the following loop, as x moves across the logical
         // horizontal pixels, we will compute a y value for each x.
         double y = y0;
         for (int x = (int)x0; x <= (int)x1; x++, y += m)
         {
            // The value of y will almost always be between
            // two vertical pixel coordinates. By rounding off
            // the value of y, we are choosing the nearest logical
            // vertical pixel coordinate.
            int x_vp = x - 1;
            int y_vp = h - (int)Math.round(y);
            if (x_vp >= 0 && x_vp < w && y_vp >= 0 && y_vp < h)
            {
               fb.setPixelVP(x_vp, y_vp, c);

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y);
            }

            // Advance (x,y) to the next pixel. Since delta_x = 1, we need delta_y = m.
         }
      }
      else  // abs(slope) > 1
      {
         if (y1 < y0) // swap (x0, y0) with (x1, y1)
         {
            double tempX = x0;
            double tempY = y0;
            x0 = x1;
            y0 = y1;
            x1 = tempX;
            y1 = tempY;
         }

         // Compute this line segment's slope.
         double m = (x1 - x0) / (y1 - y0);

         // Rasterize this line segment.
         // In the following loop, as y moves across the logical
         // vertical pixels, we will compute a x value for each y.
         double x = x0;
         for (int y = (int)y0; y <= (int)y1; x += m, y++)
         {
            // The value of x will almost always be between
            // two horizontal pixel coordinates. By rounding off
            // the value of x, we are choosing the nearest logical
            // horizontal pixel coordinate.
            int x_vp = (int)Math.round(x) - 1;
            int y_vp = h - y;
            if (x_vp >= 0 && x_vp < w && y_vp >= 0 && y_vp < h)
            {
               fb.setPixelVP(x_vp, y_vp, c);

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y);
            }

            // Advance (x,y) to the next pixel. Since delta_y = 1, we need delta_x = m.
         }
      }
   }


   private static void logPixel(int w, int h, int x, double y)
   {
      System.out.printf("[w=%d, h=%d]  x=%d, y=%.3f\n", w, h, x, y);
   }

   private static void logPixel(int w, int h, double x, int y)
   {
      System.out.printf("[w=%d, h=%d]  x=%.3f, y=%d\n", w, h, x, y);
   }
}
